/*
 * vim:ts=8:sw=3:sts=8:noexpandtab:cino=>5n-3f0^-2{2,t0,(0
 */
#include "list.h"

/* Local Structures */
typedef struct _Instance Instance;
typedef struct _Entry Entry;

typedef enum
{
   LIST_TOP,
   LIST_RIGHT,
   LIST_BOTTOM,
   LIST_LEFT,
   LIST_FLOAT
} List_Orient;

struct _Instance 
{
   Drawer_View *view;

   Evas *evas;

   Eina_List *entries;

   Evas_Object *o_box, *o_con;

   Ecore_Timer *scroll_timer;
   Ecore_Animator *scroll_animator;
   double scroll_wanted;
   double scroll_pos;

   char theme_file[4096];

   const char *parent_id;

   List_Orient orient;
};

struct _Entry
{
   Instance *inst;
   Evas_Object *o_holder, *o_icon;

   Drawer_Source_Item *si;
};

static void _list_containers_create(Instance *inst);
static Entry * _list_horizontal_entry_create(Instance *inst, Drawer_Source_Item *si);
static Entry * _list_vertical_entry_create(Instance *inst, Drawer_Source_Item *si);
static Entry * _list_horizontal_cat_create(Instance *inst, Drawer_Source_Item *si);
static Entry * _list_vertical_cat_create(Instance *inst, Drawer_Source_Item *si);
static void _list_item_pack_options(Instance *inst, Entry *e);
static void _list_autoscroll_update(Instance *inst, Evas_Coord x, Evas_Coord y, Evas_Coord w, Evas_Coord h);

static int _list_scroll_timer(void *data);
static int _list_scroll_animator(void *data);
static void _list_cb_list_mouse_move(void *data, Evas *e, Evas_Object *obj, void *event_info);
static int  _list_sort_by_category_cb(const void *d1, const void *d2);
static void _list_entry_select_cb(void *data, Evas_Object *obj, const char *emission __UNUSED__, const char *source __UNUSED__);
static void _list_entry_deselect_cb(void *data, Evas_Object *obj, const char *emission __UNUSED__, const char *source __UNUSED__);
static void _list_entry_activate_cb(void *data, Evas_Object *obj, const char *emission __UNUSED__, const char *source __UNUSED__);
static void _list_event_activate_free(void *data __UNUSED__, void *event);

EAPI Drawer_Plugin_Api drawer_plugin_api = {DRAWER_PLUGIN_API_VERSION, "List"};

EAPI void *
drawer_plugin_init(Drawer_Plugin *p, const char *id)
{
   Instance *inst = NULL;

   inst = E_NEW(Instance, 1);

   inst->view = DRAWER_VIEW(p);

   inst->parent_id = eina_stringshare_add(id);

   snprintf(inst->theme_file, sizeof(inst->theme_file),
	    "%s/e-module-drawer.edj", drawer_conf->module->dir);

   return inst;
}

EAPI int
drawer_plugin_shutdown(Drawer_Plugin *p)
{
   Instance *inst = NULL;

   inst = p->data;

   while (inst->entries)
     {
	Entry *e;

	e = inst->entries->data;
	if (e->o_icon)
	  evas_object_del(e->o_icon);
	if (e->o_holder)
	  evas_object_del(e->o_holder);
	E_FREE(e);

	inst->entries = eina_list_remove_list(inst->entries, inst->entries);
     }

   eina_stringshare_del(inst->parent_id);
   if (inst->o_box) evas_object_del(inst->o_box);
   if (inst->o_con) evas_object_del(inst->o_con);

   E_FREE(inst);

   return 1;
}

EAPI Evas_Object *
drawer_view_render(Drawer_View *v, Evas *evas, Eina_List *items)
{
   Instance *inst = NULL;
   Eina_List *l = NULL, *ll = NULL;
   Drawer_Source_Item *si;
   const char *cat = NULL;
   Eina_Bool change = EINA_FALSE;

   inst = DRAWER_PLUGIN(v)->data;

   inst->evas = evas;

   if (inst->o_box) evas_object_del(inst->o_box);
   if (inst->o_con) evas_object_del(inst->o_con);
   _list_containers_create(inst);

   e_box_freeze(inst->o_box);
   while (inst->entries)
     {
	Entry *e;

	e = inst->entries->data;
	if (e->o_icon)
	  evas_object_del(e->o_icon);
	if (e->o_holder)
	  evas_object_del(e->o_holder);
	E_FREE(e);

	inst->entries = eina_list_remove_list(inst->entries, inst->entries);
     }

   EINA_LIST_FOREACH(items, l, si)
      ll = eina_list_append(ll, si);
   ll = eina_list_sort(ll, eina_list_count(ll), _list_sort_by_category_cb);
   switch (inst->orient)
     {
      case LIST_BOTTOM:
      case LIST_RIGHT:
	 ll = eina_list_reverse(ll);
	 break;
     }

   EINA_LIST_FOREACH(ll, l, si)
     {
	Entry *e;

	if (!cat && si->category)
	  {
	     cat = eina_stringshare_add(si->category);
	     change = EINA_TRUE;
	  }
	else if (cat && !si->category)
	  {
	     eina_stringshare_del(cat);
	     cat = NULL;
	     change = EINA_TRUE;
	  }
	else if (cat && si->category && (strcmp(cat, si->category)))
	  {
	     eina_stringshare_del(cat);
	     cat = eina_stringshare_add(si->category);
	     change = EINA_TRUE;
	  }
	else
	  change = EINA_FALSE;

	switch(inst->orient)
	  {
	   case LIST_TOP:
	   case LIST_BOTTOM:
	      if (change)
		{
		   Entry *c;

		   c = _list_vertical_cat_create(inst, si);
		   inst->entries = eina_list_append(inst->entries, c);
		   e_box_pack_end(inst->o_box, c->o_holder);
		   _list_item_pack_options(inst, c);
		}
	      e = _list_vertical_entry_create(inst, si);
	      break;
	   case LIST_RIGHT:
	   case LIST_LEFT:
	   case LIST_FLOAT:
	      if (change)
		{
		   Entry *c;

		   c = _list_horizontal_cat_create(inst, si);
		   inst->entries = eina_list_append(inst->entries, c);
		   e_box_pack_end(inst->o_box, c->o_holder);
		   _list_item_pack_options(inst, c);
		}
	      e = _list_horizontal_entry_create(inst, si);
	      break;
	  }
	inst->entries = eina_list_append(inst->entries, e);
	e_box_pack_end(inst->o_box, e->o_holder);
	_list_item_pack_options(inst, e);
     }
   e_box_thaw(inst->o_box);
   eina_stringshare_del(cat);
   return inst->o_con;
}

EAPI void
drawer_view_content_size_get(Drawer_View *v, E_Gadcon_Client *gcc, Drawer_Content_Margin *margin, int *w, int *h)
{
   Instance *inst = NULL;
   Evas_Coord bw, bh, gx, gy, gw, gh, zw, zh, zx, zy;

   inst = DRAWER_PLUGIN(v)->data;
   e_box_min_size_get(inst->o_box, &bw, &bh);
   edje_extern_object_min_size_set(inst->o_box, bw, bh);
   edje_object_size_min_calc(inst->o_con, w, h);
   edje_extern_object_min_size_set(inst->o_box, 0, 0);

   e_gadcon_client_geometry_get(gcc, &gx, &gy, &gw, &gh);
   zx = gcc->gadcon->zone->x;
   zy = gcc->gadcon->zone->y;
   zw = gcc->gadcon->zone->w;
   zh = gcc->gadcon->zone->h;
   switch (gcc->gadcon->orient)
     {
      case E_GADCON_ORIENT_CORNER_RT:
      case E_GADCON_ORIENT_CORNER_RB:
      case E_GADCON_ORIENT_RIGHT:
	 if (gx - *w < zx + margin->left)
	   *w = gx - zx - margin->left;
	break;
      case E_GADCON_ORIENT_LEFT:
      case E_GADCON_ORIENT_CORNER_LT:
      case E_GADCON_ORIENT_CORNER_LB:
	if (gx + gw + *w > zx + zw + margin->right)
	  *w = zx + zw - gx - gw + margin->right;
	break;
      case E_GADCON_ORIENT_TOP:
      case E_GADCON_ORIENT_CORNER_TL:
      case E_GADCON_ORIENT_CORNER_TR:
	if (gy + gh + *h > zy + zh + margin->bottom)
	  *h = zy + zh - gy - gh + margin->bottom;
	break;
      case E_GADCON_ORIENT_BOTTOM:
      case E_GADCON_ORIENT_CORNER_BL:
      case E_GADCON_ORIENT_CORNER_BR:
	if (gy - *h < zy + margin->top)
	  *h = gy - zy - margin->top;
	break;
      case E_GADCON_ORIENT_FLOAT:
	if (*w > zw - margin->left - margin->right)
	  *w = zw - margin->left - margin->right;
	break;
     }
}

EAPI void
drawer_view_orient_set(Drawer_View *v, E_Gadcon_Orient orient)
{
   Instance *inst = NULL;
   
   inst = DRAWER_PLUGIN(v)->data;

   switch (orient)
     {
      case E_GADCON_ORIENT_CORNER_RT:
      case E_GADCON_ORIENT_CORNER_RB:
      case E_GADCON_ORIENT_RIGHT:
	e_box_orientation_set(inst->o_box, 1);
	inst->orient = LIST_RIGHT;
	break;
      case E_GADCON_ORIENT_LEFT:
      case E_GADCON_ORIENT_CORNER_LT:
      case E_GADCON_ORIENT_CORNER_LB:
	e_box_orientation_set(inst->o_box, 1);
	inst->orient = LIST_LEFT;
	break;
      case E_GADCON_ORIENT_TOP:
      case E_GADCON_ORIENT_CORNER_TL:
      case E_GADCON_ORIENT_CORNER_TR:
	e_box_orientation_set(inst->o_box, 0);
	inst->orient = LIST_TOP;
	break;
      case E_GADCON_ORIENT_BOTTOM:
      case E_GADCON_ORIENT_CORNER_BL:
      case E_GADCON_ORIENT_CORNER_BR:
	e_box_orientation_set(inst->o_box, 0);
	inst->orient = LIST_BOTTOM;
	break;
      case E_GADCON_ORIENT_FLOAT:
	e_box_orientation_set(inst->o_box, 1);
	inst->orient = LIST_FLOAT;
	break;
     }
}

static void
_list_containers_create(Instance *inst)
{
   Evas *evas;
   const char *group;

   evas = inst->evas;
   inst->o_con = edje_object_add(evas);

   inst->o_box = e_box_add(evas);
   e_box_homogenous_set(inst->o_box, 0);
   switch(inst->orient)
     {
      case LIST_TOP:
	 e_box_orientation_set(inst->o_box, 0);
	 e_box_align_set(inst->o_box, 0.5, 1);
	 group = eina_stringshare_add("modules/drawer/list/vertical");
	 break;
      case LIST_BOTTOM:
	 e_box_orientation_set(inst->o_box, 0);
	 e_box_align_set(inst->o_box, 0.5, 0);
	 group = eina_stringshare_add("modules/drawer/list/vertical");
	 break;
      case LIST_LEFT:
      case LIST_FLOAT:
	 e_box_orientation_set(inst->o_box, 1);
	 e_box_align_set(inst->o_box, 1, 0.5);
	 group = eina_stringshare_add("modules/drawer/list/horizontal");
	 break;
      case LIST_RIGHT:
	 e_box_orientation_set(inst->o_box, 1);
	 e_box_align_set(inst->o_box, 0, 0.5);
	 group = eina_stringshare_add("modules/drawer/list/horizontal");
	 break;
     }

   if (!e_theme_edje_object_set(inst->o_con, "base/theme/modules/drawer", group))
     edje_object_file_set(inst->o_con, inst->theme_file, group);

   eina_stringshare_del(group);
   edje_object_part_swallow(inst->o_con, "e.swallow.content", inst->o_box);

   evas_object_event_callback_add(inst->o_con, EVAS_CALLBACK_MOUSE_MOVE,
				  _list_cb_list_mouse_move, inst);
}

static Entry *
_list_horizontal_entry_create(Instance *inst, Drawer_Source_Item *si)
{
   Entry *e;
   Evas_Coord w, h;

   e = E_NEW(Entry, 1);

   e->o_holder = edje_object_add(inst->evas);
   if (!e_theme_edje_object_set(e->o_holder, "base/theme/modules/drawer",
				"modules/drawer/list/horizontal_item"))
     edje_object_file_set(e->o_holder, inst->theme_file,
			  "modules/drawer/list/horizontal_item");

   edje_object_part_geometry_get(e->o_holder, "e.swallow.content", NULL, NULL, &w, &h);
   e->o_icon = drawer_util_icon_create(si, inst->evas, w, h);

   edje_object_part_swallow(e->o_holder, "e.swallow.content", e->o_icon);
   evas_object_pass_events_set(e->o_icon, 1);
   evas_object_show(e->o_icon);

   e->inst = inst;
   e->si = si;

   edje_object_signal_callback_add(e->o_holder, "e,action,select", "drawer", 
				   _list_entry_select_cb, e);
   edje_object_signal_callback_add(e->o_holder, "e,action,deselect", "drawer", 
				   _list_entry_deselect_cb, e);
   edje_object_signal_callback_add(e->o_holder, "e,action,activate", "drawer", 
				   _list_entry_activate_cb, e);

   return e;
}

static Entry *
_list_vertical_entry_create(Instance *inst, Drawer_Source_Item *si)
{
   Entry *e;
   Evas_Coord w, h;

   e = E_NEW(Entry, 1);

   e->o_holder = edje_object_add(inst->evas);
   if (!e_theme_edje_object_set(e->o_holder, "base/theme/modules/drawer",
				"modules/drawer/list/vertical_item"))
     edje_object_file_set(e->o_holder, inst->theme_file, "modules/drawer/list/vertical_item");

   edje_object_part_geometry_get(e->o_holder, "e.swallow.content", NULL, NULL, &w, &h);
   e->o_icon = drawer_util_icon_create(si, inst->evas, w, h);

   edje_object_part_swallow(e->o_holder, "e.swallow.content", e->o_icon);
   evas_object_pass_events_set(e->o_icon, 1);
   evas_object_show(e->o_icon);

   edje_object_part_text_set(e->o_holder, "e.text.label", si->label);
   edje_object_part_text_set(e->o_holder, "e.text.description", si->description);

   e->inst = inst;
   e->si = si;

   edje_object_signal_callback_add(e->o_holder, "e,action,select", "drawer", 
				   _list_entry_select_cb, e);
   edje_object_signal_callback_add(e->o_holder, "e,action,deselect", "drawer", 
				   _list_entry_deselect_cb, e);
   edje_object_signal_callback_add(e->o_holder, "e,action,activate", "drawer", 
				   _list_entry_activate_cb, e);

   return e;
}

static Entry *
_list_horizontal_cat_create(Instance *inst, Drawer_Source_Item *si)
{
   Entry *e;
   char buf[1024];

   e = E_NEW(Entry, 1);

   e->o_holder = edje_object_add(inst->evas);
   if (!e_theme_edje_object_set(e->o_holder, "base/theme/modules/drawer",
				"modules/drawer/list/horizontal_category"))
     edje_object_file_set(e->o_holder, inst->theme_file,
			  "modules/drawer/list/horizontal_category");

   if (si->category)
     snprintf(buf, sizeof(buf), "%s", si->category);
   else
     snprintf(buf, sizeof(buf), "Uncategorised");

   edje_object_part_text_set(e->o_holder, "e.text.category", buf);

   e->inst = inst;
   e->si = si;

   return e;
}

static Entry *
_list_vertical_cat_create(Instance *inst, Drawer_Source_Item *si)
{
   Entry *e;
   char buf[1024];

   e = E_NEW(Entry, 1);

   e->o_holder = edje_object_add(inst->evas);
   if (!e_theme_edje_object_set(e->o_holder, "base/theme/modules/drawer",
				"modules/drawer/list/vertical_category"))
     edje_object_file_set(e->o_holder, inst->theme_file,
			  "modules/drawer/list/vertical_category");

   if (si->category)
     snprintf(buf, sizeof(buf), "%s", si->category);
   else
     snprintf(buf, sizeof(buf), "Uncategorised");

   edje_object_part_text_set(e->o_holder, "e.text.category", buf);

   e->inst = inst;
   e->si = si;

   return e;
}

static void
_list_item_pack_options(Instance *inst, Entry *e)
{
   Evas_Coord w, h, mw, mh;

   w = h = 0;
   edje_object_size_min_calc(e->o_holder, &w, &h);
   edje_object_size_max_get(e->o_holder, &mw, &mh);
   e_box_pack_options_set(e->o_holder,
			  1, 1, /* fill */
			  1, 1, /* expand */
			  0.5, 0.5, /* align */
			  w, h, /* min */
			  mw, mh /* max */
			 );
   evas_object_show(e->o_holder);
}

static void
_list_autoscroll_update(Instance *inst, Evas_Coord x, Evas_Coord y, Evas_Coord w, Evas_Coord h)
{
   double d;
   
   if (e_box_orientation_get(inst->o_box))
     {
	if (w > 1) d = (double)x / (double)(w - 1);
	else d = 0;
     }
   else
     {
	if (h > 1) d = (double)y / (double)(h - 1);
	else d = 0;
     }
   if (d < 0.0) d = 0.0;
   else if (d > 1.0) d = 1.0;
   if (!inst->scroll_timer)
     inst->scroll_timer = 
     ecore_timer_add(0.01, _list_scroll_timer, inst);
   if (!inst->scroll_animator)
     inst->scroll_animator = 
     ecore_animator_add(_list_scroll_animator, inst);
   inst->scroll_wanted = d;
}

static int
_list_scroll_timer(void *data)
{
   Instance *inst = NULL;
   double d;
   
   inst = data;
   d = inst->scroll_wanted - inst->scroll_pos;
   if (d < 0) d = -d;
   if (d < 0.001)
     {
	inst->scroll_pos =  inst->scroll_wanted;
	inst->scroll_timer = NULL;
	return 0;
     }
   inst->scroll_pos = (inst->scroll_pos * 0.95) + (inst->scroll_wanted * 0.05);
   return 1;
}

static int
_list_scroll_animator(void *data)
{
   Instance *inst = NULL;
   
   inst = data;
   if (e_box_orientation_get(inst->o_box))
     e_box_align_set(inst->o_box, 1.0 - inst->scroll_pos, 0.5);
   else
     e_box_align_set(inst->o_box, 0.5, 1.0 - inst->scroll_pos);
   if (!inst->scroll_timer)
     {
	inst->scroll_animator = NULL;
	return 0;
     }

   /* Have scroll_cb func if d&d is ever implemented. See e_gadcon.c */

   return 1;
}

static void
_list_cb_list_mouse_move(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
   Instance *inst = NULL;
   Evas_Event_Mouse_Move *ev;
   Evas_Coord x, y, w, h;
   
   ev = event_info;
   inst = data;
   evas_object_geometry_get(inst->o_box, &x, &y, &w, &h);
   _list_autoscroll_update(inst, ev->cur.output.x - x, 
			   ev->cur.output.y - y, w, h);
}

static int
_list_sort_by_category_cb(const void *d1, const void *d2)
{
   const Drawer_Source_Item *si1;
   const Drawer_Source_Item *si2;
   int ret;

   if (!(si1 = d1)) return -1;
   if (!(si2 = d2)) return 1;

   if (!si1->category) return -1;
   if (!si2->category) return 1;

   ret = strcmp(si1->category, si2->category);
   
   return ret > 0 ? 1 : -1;
}

static void
_list_entry_select_cb(void *data, Evas_Object *obj, const char *emission __UNUSED__, const char *source __UNUSED__)
{
   Entry *e = NULL;
   Instance *inst = NULL;
   Drawer_Source_Item *si = NULL;

   e = data;
   inst = e->inst;
   si = e->si;
   edje_object_part_text_set(inst->o_con, "e.text.label", si->label);
   edje_object_part_text_set(inst->o_con, "e.text.description", si->description);
}

static void
_list_entry_deselect_cb(void *data, Evas_Object *obj, const char *emission __UNUSED__, const char *source __UNUSED__)
{
   Entry *e = NULL;
   Instance *inst = NULL;

   e = data;
   inst = e->inst;
   edje_object_part_text_set(inst->o_con, "e.text.label", NULL);
   edje_object_part_text_set(inst->o_con, "e.text.description", NULL);
}

static void
_list_entry_activate_cb(void *data, Evas_Object *obj, const char *emission __UNUSED__, const char *source __UNUSED__)
{
   Entry *e = NULL;
   Drawer_Event_View_Activate *ev;

   e = data;
   ev = E_NEW(Drawer_Event_View_Activate, 1);
   ev->data = e->si;
   ev->view = e->inst->view;
   ev->id = eina_stringshare_add(e->inst->parent_id);
   ecore_event_add(DRAWER_EVENT_VIEW_ITEM_ACTIVATE, ev, _list_event_activate_free, NULL);

   /* XXX: this doesn't seem to work */
   edje_object_signal_emit(e->inst->o_con, "e,action,activate", "drawer");
}

static void
_list_event_activate_free(void *data __UNUSED__, void *event)
{
   Drawer_Event_View_Activate *ev;

   ev = event;
   eina_stringshare_del(ev->id);
   free(ev);
}
